package gfcutil;

import GFC.*;
import GFC.Absyn.*;
import util.*;

import java.util.*;
import java.io.*;

/**
 *  Contains a collection of GFC modules which share an abstract syntax.
 */
public class GFCMModules {

        /**
         *  Maps module names to modules.
         */
        private Map<String,GFCModule> map;

        /**
         *  The name of the main abstract module.
         */
        private String mainAbstract;

        public GFCMModules(Canon gfc) {
                this.map = new TreeMap<String,GFCModule>();
                this.mainAbstract = null;

                Gr gr = (Gr)gfc;
                for (GFC.Absyn.Module m : gr.listmodule_) {
                        GFCModule gm = new GFCModule((Mod)m, map);
                        add(gm);

//                        System.err.println("Added " + gm.getType() + " module " + gm.getName() + " of " + gm.getAbstract());

                        if (gm.getType() == ModuleType.ABSTRACT && mainAbstract == null)
                                mainAbstract = gm.getName();
                }
        }

        /**
         *  Gets a module by name.
         */
        public GFCModule getModule(String name) {
                return map.get(name);
        }

        /**
         *  Gets the main abstract module.
         */
        public GFCModule getMainAbstractModule() {
                return getModule(mainAbstract);
        }

        /**
         *  Gets the main concrete modules, i.e. those that are concrete syntaxes
         *  for the main abstract module.
         */
        public List<GFCModule> getMainConcreteModules() {
                List<GFCModule> ms = new LinkedList<GFCModule>();
                for (GFCModule m : map.values())
                        if (m.getType() == ModuleType.CONCRETE 
                            && m.getAbstract().equals(mainAbstract))
                                ms.add(m);
                return ms;
        }

        /**
         *  Add a GFCModule to this collection of modules.
         */
        private void add(GFCModule module) {
                String name = module.getName();
                if (map.containsKey(name))
                        throw new RuntimeException("Module " 
                                                   + name + " already exists.");
                                
                map.put(name, module);
        }

        /**
         *  Load a GFCM grammar from a file.
         *
         *  @param file Filename.
         */
        public static GFCMModules loadGFCM(String file) 
                throws FileNotFoundException, IOException {

                return loadGFCM(FileUtil.openFile(file));
        }

        /**
         *  Load a GFCM grammar from a Reader.
         */
        public static GFCMModules loadGFCM(Reader input) throws IOException {
                System.err.println("Parsing GFCM file");
                Canon gfc = GFCParser.parseCanon(input);
                System.err.println("Done parsing GFCM file");
                input.close();
                return new GFCMModules(gfc);
        }

        /**
         *  Load a gzipped GFCM grammar from a file.
         */
/*
        public static GFCMModules loadGFCM_GZ(String file) throws FileNotFoundException, IOException {
                return loadGFCM(new java.util.zip.GZIPInputStream(new FileInputStream(file)));
        }
*/
}
