package gfcutil;

import abssyn.*;
import GFC.*;
import GFC.Absyn.*;
import java.util.*;
import java.io.Serializable;

/**
 *  Wraps a GFC module (concrete, abstract or resource) 
 *  and provides methods for looking up its contents.
 */
public class GFCModule implements Serializable {

        /** Module name. */
        private String name;

        /** Ttype of this module. */
        private ModuleType type;

        /** 
         *  The abstract module that this module is for. null
         *  if this module is not concrete.
         */
        private String abstractModule;

        /** Maps function names to linearization rules. */
        private Map<String,LinRule> rules;

        /** Maps function names to functions. */
        private Map<String,Function> funs;

        /** 
         *  Maps all inherited names to the modules that they are 
         *  inherited from.
         */
        private Map<String,Inherited> inherited;

        /** Maps module names to modules. */
        private Map<String,GFCModule> allModules;


        /**
         *  @param mod A module parse tree
         *  @param scope The names of the modules in scope in the 
         *    current module.
         *  @param allModules A reference to a map of module names to 
         *    modules. The map need not be populated when passed to 
         *    the contructor, but it must be before find is called.
         */
        public GFCModule(Mod mod, /*List<String> scope,*/ Map<String,GFCModule> allModules) {
                this.name = mod.modtype_.accept(new GetModName(), null);
                this.type = ModuleType.getModuleType(mod.modtype_);
                this.abstractModule = getAbstractModule(mod);
                this.rules = getDefs(mod, new GetLinRule());
                this.funs = getDefs(mod, new GetFunction());
                this.inherited = getDefs(mod, new GetInherited());
                this.allModules = allModules;
        }

        /**
         *  Get the name of this module.
         */
        public String getName() {
                return name;
        }

        /**
         *  Checks if this module is concrete.
         */
        public ModuleType getType() {
                return type;
        }

        /**
         *  Get the name of the abstract module that this module is for.
         *  If this is not a concrete module, null is returned.
         */
        public String getAbstract() {
                return abstractModule;
        }

        /**
         *  Find a function by name. 
         */
        public Function findFunction(String f) {
                System.err.println(name + ".findFunction("+f+")");

                Function fun = funs.get(f);
                if (fun != null) 
                        return fun;
                Inherited in = inherited.get(f);
                if (in == null)
                        throw new RuntimeException("No function called " + f + " in module " + name);
                GFCModule parent = allModules.get(in.getModule());
                return parent.findFunction(f);
        }

        /**
         *  Find a linearization rule by function name.
         */
        public LinRule findLinRule(String f) {
                LinRule rule = rules.get(f);
                if (rule != null) 
                        return rule;
                Inherited in = inherited.get(f);
                if (in == null)
                        throw new RuntimeException("No linearization rule for " + f);
                GFCModule parent = allModules.get(in.getModule());
                return parent.findLinRule(f);
        }

        public boolean equals(Object o) {
                return o instanceof GFCModule && equals((GFCModule)o);
        }

        public boolean equals(GFCModule mod) {
                return getName().equals(mod.getName());
        }

        public int hashCode() {
                return getName().hashCode();
        }

        /**
         *  Gets the abstract module that a concrete module is for.
         */
        private static String getAbstractModule(Mod mod) {
                List<String> as = mod.modtype_.accept(new GetAbstractModule(), new LinkedList<String>());
                if (as.size() == 1)
                        return as.get(0);
                assert as.size() == 0;
                return null;
        }

        /**
         *  Gets the abstract module that a concrete module is for.
         */
        private static class GetAbstractModule implements ModType.Visitor<List<String>,List<String>> { 
                public List<String> visit(GFC.Absyn.MTAbs p, List<String> arg) { return arg; }
                public List<String> visit(GFC.Absyn.MTCnc p, List<String> arg) { 
                        arg.add(p.ident_2);
                        return arg; 
                }
                public List<String> visit(GFC.Absyn.MTRes p, List<String> arg) { return arg; }
                public List<String> visit(GFC.Absyn.MTTrans p, List<String> arg) { return arg; }
        }

        /**
         *  Gets all modules opened by a module;
         */
        private static class GetOpenModules implements Open.Visitor<List<String>,List<String>> { 
                public List<String> visit(Opens p, List<String> arg) { 
                        arg.addAll(p.listident_);
                        return arg;
                }
                public List<String> visit(NoOpens p, List<String> arg) { 
                        return arg;
                }
        }

        /**
         *  Gets the module (if there is one) that a module extends.
         */
        private static class GetExtendedModule implements Extend.Visitor<List<String>,List<String>> {
                public List<String> visit(Ext p, List<String> arg) {
                        arg.add(p.ident_);
                        return arg;
                }
                public List<String> visit(NoExt p, List<String> arg) {
                        return arg;
                }
        }


        /**
         *  Gets certain definitions from a module.
         */
        private static <A extends Definition> Map<String,A> getDefs(Mod mod, GetDef<A> v) {
                return mod.accept(new GetDefs<A>(v), new HashMap<String, A>());
        }

        /**
         *  Uses a GetDef on a ll defintions in a module.
         */
        private static class GetDefs<A extends Definition> implements Module.Visitor<Map<String,A>,Map<String,A>> {
                private GetDef<A> v;
                public GetDefs(GetDef<A> v) { this.v = v; }
                public Map<String,A> visit(GFC.Absyn.Mod mod, Map<String,A> arg) {
                        for (GFC.Absyn.Def d : mod.listdef_) {
                                A r = d.accept(v, arg);
                                if (r != null) {
                                        String name = r.getName();
                                        if (arg.containsKey(name))
                                                throw new RuntimeException(
                                                        name + " already exists");
                                        arg.put(name, r);
                                }
                        }
                        return arg;
                }
        }

        /**
         *  Visits a GFC ModType and gets the module name.
         */
        private static class GetModName implements ModType.Visitor<String,Object> {
                public String visit(GFC.Absyn.MTAbs p, Object arg) {
                        return p.ident_;
                }
                public String visit(GFC.Absyn.MTCnc p, Object arg) {
                        return p.ident_1;
                }
                public String visit(GFC.Absyn.MTRes p, Object arg) {
                        return p.ident_;
                }
                public String visit(GFC.Absyn.MTTrans p, Object arg) {
                        return p.ident_1;
                }
        }

        /**
         *  Base class for GFC definition visitors. Returns null for all definitions.
         */
        private static abstract class GetDef<A> implements Def.Visitor<A,Object> {
                public A visit(GFC.Absyn.CncDFun cncdfun, Object arg) { return null; }
                public A visit(GFC.Absyn.AbsDCat p, Object arg) { return null; }
                public A visit(GFC.Absyn.AbsDFun p, Object arg) { return null; }
                public A visit(GFC.Absyn.AbsDTrans p, Object arg) { return null; }
                public A visit(GFC.Absyn.ResDPar p, Object arg) { return null; }
                public A visit(GFC.Absyn.ResDOper p, Object arg) { return null; }
                public A visit(GFC.Absyn.CncDCat p, Object arg) { return null; }
                public A visit(GFC.Absyn.AnyDInd p, Object arg) { return null; }
        }

        /**
         *  Visits a GFC definition. If it is a linearization rule definition, 
         *  a LinRule is returned, otherwise null is returned.
         */
        private static class GetLinRule extends GetDef<LinRule> {
                public LinRule visit(GFC.Absyn.CncDFun cncdfun, Object arg) {
                        return new LinRule(cncdfun.ident_, 
                                           cncdfun.listargvar_, 
                                           cncdfun.term_1);
                }
        }

        /**
         *  Visits a GFC definition. If it is a function definition, 
         *  a Function is returned, otherwise null is returned.
         */
        private static class GetFunction extends GetDef<Function> {
                 public Function visit(GFC.Absyn.AbsDFun p, Object arg) {
                         // FIXME: is this always true?
                         if (!(p.exp_2 instanceof EData))
                                 throw new RuntimeException("p.exp_2 is not data");

                         return new Function(p.ident_, p.exp_1);
                 }
        }

        /**
         *  Visits a GFC definition. If it is a inheritance definition, 
         *  an Inherited is returned, otherwise null is returned.
         */
        private static class GetInherited extends GetDef<Inherited> {
                 public Inherited visit(GFC.Absyn.AnyDInd p, Object arg) {
                         return new Inherited(p.ident_1, p.ident_2);
                 }
        }

}
