package linearize;

import GFC.*;
import GFC.Absyn.*;
import linearize.gfvalue.*;
import util.Pair;

import java.util.*;

import static linearize.gfvalue.Concat.concat;


/**
 *  Does linearization by interpreting a GFC linearization rule.
 */
public class InterpretLinRule implements Term.Visitor<Value,Value[]> {

        private static Tokn.Visitor<Value,Object> toknToValue 
                = new Tokn.Visitor<Value,Object>() {
                        public Value visit(GFC.Absyn.KS p, Object arg) {
                                return new Str(p.string_);
                        }

                        /* Prefix-dependent choice */
                        // KP.   Tokn  ::= "[" "pre" [String] "{" [Variant] "}" "]" ;
                        // Var.  Variant ::= [String] "/" [String] ;
                        // default, then value / prefix?
                        public Value visit(GFC.Absyn.KP p, Object arg) {
                                Value def = concat(p.liststring_);

                                List<Pair<Value,List<String>>> vs = 
                                  new LinkedList<Pair<Value,List<String>>>();
                                for (Variant var : p.listvariant_) {
                                        // FIXME: use visitor
                                        Var v = (Var)var;
                                        Value t = concat(v.liststring_1);
                                        vs.add(new Pair<Value,List<String>>(t, v.liststring_2));
                                }

                                return new Pre(def, vs);
                        }
                };

        private static Label.Visitor<String,Object> labelToString 
                = new Label.Visitor<String,Object>() {
                        public String visit(GFC.Absyn.L p, Object arg) {
                                return p.ident_;
                        }

                        // FIXME: implement
                        // LV.   Label  ::= "$" Integer ;
                        // Doesn't seem to occur in canonical GF grammars?
                        public String visit(GFC.Absyn.LV p, Object arg) {
                                 return unused(this, p, arg); 
                        }
                };

        private static String cidentToString(CIdent cid) {
                return cid.accept(new CIdent.Visitor<String,Object>() {
                        public String visit(GFC.Absyn.CIQ p, Object arg) {
                                return p.ident_1 + "." + p.ident_2;
                        }}, null);
        }

        private static ArgVar.Visitor<Value,Value[]> argVarValue
                = new ArgVar.Visitor<Value,Value[]>() {

                        /* simple argument variable */
                        public Value visit(GFC.Absyn.A p, Value[] arg) {
                                return arg[p.integer_];
                        }
                        

                        // FIXME: implement 
                        // What is this?
                        // AB.   ArgVar ::= Ident "+" Integer "@" Integer ; -- with a number of bindings
                        public Value visit(GFC.Absyn.AB p, Value[] arg) {
                                //p.ident_;
                                //p.integer_1;
                                //p.integer_2;
                                return unused(this, p, arg);
                        }
                };

        private static Patt.Visitor<Value,Object> patternValue 
        = new Patt.Visitor<Value,Object>() {
                public Value visit(PC p, Object arg) {
                        String s = cidentToString(p.cident_);
                        Value[] ps = new Value[p.listpatt_.size()];
                        int i = 0;
                        for (Patt pa : p.listpatt_)
                                ps[i++] = pa.accept(this, arg);
                        return new Param(s, ps);
                }
                public Value visit(GFC.Absyn.PV p, Object arg) {
                        return new Param(p.ident_, new Param[0]);
                }
                
                /* wildcard */
                // FIXME: implement
                // should produce multiple params
                public Value visit(GFC.Absyn.PW p, Object arg) {
                        return unused(this, p, arg); // FIXME
                }

                /* Pattern matching on records */
                // FIXME: implement
                // PR.   Patt ::= "{" [PattAssign] "}" ;
                // PAss. PattAssign ::= Label "=" Patt ;
                public Value visit(GFC.Absyn.PR p, Object arg) {
                        Record r = new Record();
                        for (PattAssign pass : p.listpattassign_) {
                                PAss pa = (PAss)pass; // FIXME: use visitor?
                                String l = pa.label_.accept(labelToString, null);
                                Value v = pa.patt_.accept(this, arg);
                                r.set(l, v);
                        }
                        return r;
                }
        };

        private <X extends Term> List<Value> visitList(List<X> xs, Value[] arg) {
                List<Value> ys = new LinkedList<Value>();
                for (X x : xs)
                        ys.add(x.accept(this, arg));
                return ys;
        }

        /* record construction */
        public Value visit(GFC.Absyn.R p, Value[] arg) {
                Record r = new Record();
                for (Assign a : p.listassign_) {
                        // FIXME: use Assign.Visitor
                        Ass ass = (Ass)a;
                        String l = ass.label_.accept(labelToString, null);
                        r.set(l, ass.term_.accept(this, arg));
                }
                return r;
        }

        /* concatenation */
        public Value visit(GFC.Absyn.C p, Value[] arg) {
                Value v1 = p.term_1.accept(this, arg);
                Value v2 = p.term_2.accept(this, arg);
                return new Concat(v1,v2);
        }

        /* string constant */
        public Value visit(GFC.Absyn.K p, Value[] arg) {
                return p.tokn_.accept(toknToValue, null);
        }
        
        /* record projection  */
        public Value visit(GFC.Absyn.P p, Value[] arg) {
                Value v = p.term_.accept(this, arg);
                String l = p.label_.accept(labelToString, null);
                if (!(v instanceof Record))
                        throw new RuntimeException(v.getClass().getName() + " v = '" + v + "' is not a record in projection v." + l);
                Record r = (Record)v;
                return r.get(l);
        }

        /* argument variable */
        public Value visit(GFC.Absyn.Arg a, Value[] arg) {
                return a.argvar_.accept(argVarValue, arg);
        }



        /* parameter constructor */
        public Value visit(GFC.Absyn.Con con, Value[] arg) {
                Value[] as = visitList(con.listterm_, arg).toArray(new Value[con.listterm_.size()]);
                return new Param(cidentToString(con.cident_), as);
        }

        /* table selection */
        public Value visit(GFC.Absyn.S s, Value[] arg) {
                Value v = s.term_1.accept(this, arg);
                if (!(v instanceof linearize.gfvalue.Table))
                        throw new RuntimeException(v.getClass().getName() 
                                                   + " " + v 
                                                   + " is not a table");

                 linearize.gfvalue.Table t = 
                         (linearize.gfvalue.Table)s.term_1.accept(this, arg);
                 Value p = s.term_2.accept(this, arg);
                 return t.select(p);
        }

        /* table construction */
        public Value visit(GFC.Absyn.T t, Value[] arg) {
//                t.ctype_.accept(this, arg);
                Map<Value,Value> tm = new HashMap<Value,Value>();
                for (Case c : t.listcase_) {
                        Cas k = (Cas)c; // FIXME
                        Value v = k.term_.accept(this, arg);
                        for (Patt p : k.listpatt_) {
                                Value pa = p.accept(patternValue, null);
                                tm.put(pa, v);
                        }
                }
                return new linearize.gfvalue.Table(tm);
        }

        /* empty */
        public Value visit(GFC.Absyn.E p, Value[] arg) {
                return new Str(""); // FIXME: { s = ... } ?
        }

        /** Free variation. We just pick the first one if there is one. */
        public Value visit(GFC.Absyn.FV p, Value[] arg) {
                if (p.listterm_.size() == 0) {
                        //       throw new RuntimeException("No variants.");
                        return new Str("<no variants>");
                }
                return p.listterm_.get(0).accept(this, arg);
        }

        // FIXME: implement
        // LI.   Term2 ::= "$" Ident ;  -- from pattern variables
        public Value visit(GFC.Absyn.LI p, Value[] arg) {
                return unused(this, p, arg);
        }

        // FIXME: implement
        // I.    Term2 ::= CIdent ; -- from resources
        public Value visit(GFC.Absyn.I p, Value[] arg) {
                return unused(this, p, arg);
        }

        private static <R,A> R unused(Object v, Object p, A arg) {
                throw new RuntimeException(v.getClass().getName() + " has no case for " + p.getClass().getName());
        }

}
