package parse.chart;

import abssyn.*;

import java.util.*;

public class TreeBuilder {

        private Chart chart;

        public TreeBuilder(Chart chart) {
                this.chart = chart;
        }

        /**
         *  Get all trees for the given category spanning the entire input.
         */
        public List<Tree> buildTree(Category cat) {
                return buildTree(cat, 0, chart.getSize());
        }

        /**
         *  Get all trees for the given category between nodes i and j.
         */
        private List<Tree> buildTree(Category cat, int i, int j) {
//                System.err.println("Building tree for category '" + cat + "' between " + i + " and " + j);

                assert i <= j : i + ", " + j;

                List<Tree> ts = new LinkedList<Tree>();

                for (Edge e : chart.get(i, j)) {
                        if (e.getCategory().equals(cat)) {
//                                System.err.println("Considering " + e);
                                Rule r = e.getRule();
                                // FIXME: unelegant to use type case here
                                if (r instanceof EmptyRule) {
                                        EmptyRule er = (EmptyRule)r;
                                        Tree[] children = new Tree[er.getArgCount()];
                                        for (int x = 0; x < children.length; x++)
                                                children[x] = new MetaVariable();
                                        ts.add(new Fun(r.getFun(), children));
                                } else if (r instanceof TerminalRule) {
                                        ts.add(new Fun(r.getFun(), 0));
                                } else {
                                        NonTerminalRule nr = (NonTerminalRule)r;
                                        List<List<Tree>> css = buildTrees(nr.getRhs(), i, j);
                                        for (List<Tree> cs : css) {
                                                Tree[] children = keepNonTerminals(cs);
                                                Tree t = makeNode(nr, children);
                                                if (t != null)
                                                        ts.add(t);
                                        }
                                }
                        }
                }
         
                return ts;
        }

        private Tree[] keepNonTerminals(List<Tree> ts) {
                List<Tree> r = new ArrayList<Tree>(ts.size());
                for (Tree t : ts)
                        if (treeIsNonTerminal(t))
                                r.add(t);
                return r.toArray(new Tree[r.size()]);
        }

        // FIXME: hackish
        private boolean treeIsNonTerminal(Tree t) {
                return t instanceof Fun && !((Fun)t).getLabel().startsWith("$");
        }

        /**
         *  Make an abstract syntax node for a rule given the trees for the children.
         *
         *  @param children Child trees from context free rule
         *  @return A syntax tree node, or null if the children could not be unified
         *          according to the rule profile.
         */
        private Tree makeNode(NonTerminalRule r, Tree[] children) {
                int[][] profile = r.getProfile();
                
                // child trees used to construct the node
                Tree[] childNodes = new Tree[profile.length];
                for (int i = 0; i < childNodes.length; i++) {
                        childNodes[i] = new MetaVariable();
                        for (int x : profile[i]) {
                                childNodes[i] = unifyTrees(childNodes[i], children[x]);
                                if (childNodes[i] == null)
                                        return null;
                        }
                }

                return new Fun(r.getFun(), childNodes);
        }

        public static Tree unifyTrees(Tree t1, Tree t2) {
                if (t1 == null || t2 == null)
                        throw new NullPointerException();

                Tree t = t1.accept(new TreeUnifier(), t2);
//                if (t == null)
//                        System.err.println("Can't unify " + t1 + " and " + t2);

//                System.err.println("unifyTrees(" + t1 + ", " + t2 + ") => " + t);

                return t;
        }

        // FIXME: add diagnostic messages when it fails
        private static class TreeUnifier implements TreeVisitor<Tree,Tree> {
                public Tree visit(Fun f, Tree t) {
                        if (t instanceof MetaVariable)
                                return f;
                        if (!(t instanceof Fun))
                                return null;
                        Fun g = (Fun)t;
                        if (!f.getLabel().equals(g.getLabel()))
                                return null;
                        Tree[] cs1 = f.getChildren();
                        Tree[] cs2 = g.getChildren();
                        if (cs1.length != cs2.length)
                                return null;
                        Tree[] cs = new Tree[cs1.length];
                        for (int i = 0; i < cs.length; i++) {
                                Tree c = unifyTrees(cs1[i], cs2[i]);
                                if (c == null)
                                        return null;
                                cs[i] = c;
                        }
                        return new Fun(f.getLabel(), cs);
                }
                public Tree visit(Literal l, Tree t) {
                        if (!l.equals(t)) {
                                return null;
                        }
                        return l;
                }
                public Tree visit(MetaVariable v, Tree t) {
                        return t;
                }
        }


        /**
         *  Get all tree lists for the given symbols between i and j.
         *  Each element in the result is one of the ambigous parses. Each such element
         *  is a list of trees that together cover the input from i to j and where
         *  each tree corresponds to one of the given symbols.
         */
        private List<List<Tree>> buildTrees(List<Category> ss, int i, int j) {
                assert ss.size() > 0;
                assert i <= j : i + ", " + j;

                List<List<Tree>> r = new LinkedList<List<Tree>>();

                // head is the first category that we are looking for
                Category head = ss.get(0);
                // tail is the rest of the categories
                List<Category> tail = ss.subList(1, ss.size());

                // go through all possible first edges
                // k is the end index of the first edge
                for (int k = i; k <= j; k++) {
                        // head is the last category that we are looking for, 
                        // but we are not at the end of the input
                        // keep looping until k = j
                        if (tail.isEmpty() && k < j)
                                continue;

                        // go through all first edges of the right category
                        for (Edge e : chart.get(i, k)) {
                                if (!e.getCategory().equals(head))
                                        continue;

                                // build all possible trees for the first category
                                List<Tree> hts = buildTree(head, i, k);
                                // build all possible tree lists for tail
                                if (tail.isEmpty()) {
                                        assert k == j : k + ", " + j;
                                        for (Tree h : hts) {
                                                List<Tree> ts = new LinkedList<Tree>();
                                                ts.add(h);
                                                r.add(ts);
                                        }
                                } else {
                                        List<List<Tree>> rss = buildTrees(tail, k, j);
                                        // make all combinations of head tree and the rest of the trees.
                                        for (Tree h : hts) {
                                                for (List<Tree> rs : rss) {
                                                        List<Tree> ts = new LinkedList<Tree>(rs);
                                                        ts.add(0, h);
                                                        r.add(ts);
                                                }
                                        }
                                }
                        }
                }

                return r;
        }
                        
}
