package translate;

import gfcutil.*;
import parse.*;
import parse.chart.*;
import linearize.*;

import java.io.*;
import java.util.*;


/**
 *  Loads parsers, linearizers etc. and creates Translators from them.
 */
public class TranslatorFactory {

        private static String getProp(Properties prop, String name) {
                String value = prop.getProperty(name);
                if (value == null)
                        throw new RuntimeException("Property " + name + " is not set");
                return value;
        }

        private static String[] splitAtCommas(String in) {
                if (in == null)
                        return new String[0];
                String[] ss = in.split("\\s*,\\s*");
                List<String> l = new ArrayList<String>(ss.length);
                for (String s : ss)
                        if (s.length() > 0)
                                l.add(s);
                return l.toArray(new String[l.size()]);
        }

        private static List<Object> instantiateAll(String[] classes) {
                List<Object> cs = new LinkedList<Object>();
                for (String n : classes) {
                        try {
                                Class<?> c = Class.forName(n);
                                cs.add(c.newInstance());
                        } catch (ClassNotFoundException ex) {
                                System.err.println(ex.toString());
                        } catch (InstantiationException ex) {
                                System.err.println(ex.toString());
                        } catch (IllegalAccessException ex) {
                                System.err.println(ex.toString());
                        }
                }
                return cs;
        }

        private static Reader readResource(String name) throws IOException {
                return readResource(name, "UTF-8");
        }

        private static Reader readResource(String name, String charset) throws IOException {
                InputStream in = Translator.class.getResourceAsStream(name);
                if (in == null)
                        throw new FileNotFoundException("Resource " + name + " not found.");
                return new InputStreamReader(in, charset);
        }

        /**
         *  Create a translator by reading the resource "/translate.properties".
         */
        public static Translator createTranslator () throws IOException {
                return createTranslator("/translate.properties");
        }

        /**
         *  Create a translator by reading given resource.
         */
        public static Translator createTranslator (String propFile) throws IOException {
                Properties prop = new Properties();
                InputStream propIn = Translator.class.getResourceAsStream(propFile);
                prop.load(propIn);
                propIn.close();

                String cfgm = getProp(prop, "cfgm");
                String[] xps = splitAtCommas(prop.getProperty("extraparsers"));
                String gfcm = getProp(prop, "gfcm");
                String[] xls = splitAtCommas(prop.getProperty("extralinearizers"));

                return createTranslator(cfgm, gfcm, xps, xls);
        }

        public static Translator createTranslator (String cfgm, 
                                                   String gfcm, 
                                                   String[] extraparsers, 
                                                   String[] extralinearizers) 
                throws IOException {

                System.err.println("Creating parsers from " + cfgm + "...");
                Parsers ps = CFGParsers.readParsers(readResource("/"+cfgm));

                ps.addParser(new ParseAll(ps));

                for (Object o : instantiateAll(extraparsers)) {
                        Parser p = (Parser)o;
                        System.err.println("Loaded extra parser " + p.getName());
                        ps.addParser(p);
                }

                System.err.println("Creating linearizers from " + gfcm + "...");
                GFCMModules ms = GFCMModules.loadGFCM(readResource("/"+gfcm));
                Linearizers ls = new GFCMLinearizers(ms);

                ls.addLinearizer(new LinearizeAll(ls));

                for (Object o : instantiateAll(extralinearizers)) {
                        Linearizer l = (Linearizer)o;
                        System.err.println("Loaded extra linearizer " + l.getName());
                        ls.addLinearizer(l);
                }

                // FIXME: this is not neccessarily the right module, we just pick the first one
                GFCModule absModule = ms.getMainAbstractModule();
                TypeAnnotator annotator = new TypeAnnotator(absModule);

                return new Translator(ps, annotator, ls);
        }

}
